from typing import List

from coiote.utils import ApiEndpoint, api_call
from coiote.v3.model.data_model import DeviceDataEntry


class DataModel(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="cachedDataModels")

    @api_call(List[DeviceDataEntry])
    def get_device_datamodel(self, device_id: str, parameters=None):
        if parameters is None:
            parameters = []
        return self.session.get(self.get_url(f"/{device_id}"), params={"parameters": parameters})

    def get_device_parameter(self, device_id: str, parameter_key: str) -> List[DeviceDataEntry]:
        return self.get_device_datamodel(device_id, [parameter_key])



from dataclasses import dataclass, field
from typing import List, Dict


@dataclass
class DeviceTestSchedule:
    testCases: List[str] = field(default_factory=list)


@dataclass
class TestExecutionReport:
    passedSuccessfully: List[str] = field(default_factory=list)
    waitingForExecution: List[str] = field(default_factory=list)
    passedWithWarning: Dict[str, str] = field(default_factory=dict)
    failed: Dict[str, str] = field(default_factory=dict)


from dataclasses import asdict
from datetime import datetime
from typing import Optional
import pytz

from requests import Response

from coiote.utils import ApiEndpoint, api_call, ISO_INSTANT_FORMAT, sanitize_request_param, api_call_raw
from coiote.v3.model.device_monitoring import MonitoringStatus, MonitoringDataResponse, SetResourceAliasRequest


class DeviceMonitoring(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="deviceMonitoring")

    @api_call_raw
    def set_resource_alias(self, device_id: str, resource: str, alias: str):
        device_id = sanitize_request_param(device_id)
        body = SetResourceAliasRequest(lwm2mUrl=resource, alias=alias)
        return self.session.post(self.get_url(f"/configuration/device/{device_id}/alias"), json=asdict(body))

    @api_call_raw
    def delete_resource_alias(self, device_id: str, resource: str):
        device_id = sanitize_request_param(device_id)
        resource = sanitize_request_param(resource)
        return self.session.delete(self.get_url(f"/configuration/device/{device_id}/resourceUrl/{resource}"))

    @api_call_raw
    def get_aliases_for_device(self, device_id: str) -> Response:
        device_id = sanitize_request_param(device_id)
        return self.session.get(self.get_url(f"/configuration/device/{device_id}/aliases"))

    @api_call_raw
    def toggle_monitoring_for_device(self, device_id: str, enabled: bool = True):
        device_id = sanitize_request_param(device_id)
        if enabled:
            url = self.get_url(f"/configuration/device/{device_id}/enable")
        else:
            url = self.get_url(f"/configuration/device/{device_id}/disable")
        return self.session.post(url)

    @api_call_raw
    def toggle_monitoring_for_group(self, group_id: str, enabled: bool = True):
        group_id = sanitize_request_param(group_id)
        if enabled:
            url = self.get_url(f"/configuration/group/{group_id}/enable")
        else:
            url = self.get_url(f"/configuration/group/{group_id}/disable")
        return self.session.post(url)

    @api_call(MonitoringStatus)
    def is_device_monitored(self, device_id: str) -> Response:
        return self.session.get(self.get_url(f"/configuration/device/{device_id}/enabled"))

    @api_call(MonitoringDataResponse)
    def get_data_batch(self,
                       device_id: str,
                       lwm2m_url: Optional[str] = None,
                       alias: Optional[str] = None,
                       start_time: Optional[datetime] = None,
                       end_time: Optional[datetime] = None,
                       limit=2048
                       ) -> Response:
        if lwm2m_url is not None:
            safe_lwm2m_url = sanitize_request_param(lwm2m_url)
            url = self.get_url(f"/data/{device_id}/resourceUrl/{safe_lwm2m_url}")
        elif alias is not None:
            safe_alias = sanitize_request_param(alias)
            url = self.get_url(f"/data/{device_id}/alias/{safe_alias}")
        else:
            raise ValueError("You must specify either LwM2M URL or alias of the resource")
        if start_time is None:
            raise ValueError("You must specify at least the start time for the query")
        params = {"timeRangeStart": start_time.astimezone(pytz.utc).strftime(ISO_INSTANT_FORMAT), "limit": limit}
        if end_time:
            params["timeRangeEnd"] = end_time.astimezone(pytz.utc).strftime(ISO_INSTANT_FORMAT)
        return self.session.get(url, params=params)



from dataclasses import asdict

from requests import Response

from coiote.utils import ApiEndpoint, api_call, sanitize_request_param, api_call_raw
from coiote.v3.model.tasks import *


class Tasks(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="tasks")

    @api_call(str)
    def run_device_config_task(self, device_id: str, task_definition: ConfigurationTaskDefinition) -> Response:
        return self.session.post(self.get_url(f"/configure/{device_id}"),
                                 json={'taskDefinition': asdict(task_definition)})

    @api_call(List[str])
    def get_all(self, search_criteria: Optional[str] = None) -> Response:
        params = {"searchCriteria": search_criteria}
        return self.session.get(self.get_url(), params=params)

    @api_call_raw
    def delete_callback(self, task_id: str, callback_id: str):
        task_id = sanitize_request_param(task_id)
        callback_id = sanitize_request_param(callback_id)
        return self.session.delete(self.get_url(f"/callback/{task_id}/{callback_id}"))

    @api_call(str)
    def run_device_fota_task(self,
                             device_id: str,
                             firmware_resource_id: str,
                             transfer_method: TransferMethod = TransferMethod.Pull,
                             transfer_protocol: TransferProtocol = TransferProtocol.HTTP,
                             timeout: str = "20m",
                             use_quota: bool = True,
                             upgrade_strategy: UpgradeStrategy = UpgradeStrategy.WithoutObservations,
                             blocking: bool = True,
                             use_cache_for_initial_state_read: bool = False,
                             check_delivery_and_protocol: bool = True,
                             resume_after_downlink_failure: bool = False,
                             execute_immediately: bool = False
                             ) -> Response:
        params = {
            "executeImmediately": execute_immediately,
            "resumeAfterDownlinkFailure": resume_after_downlink_failure,
            "checkDeliveryAndProtocol": check_delivery_and_protocol,
            "useCacheForInitialStateRead": use_cache_for_initial_state_read,
            "blocking": blocking,
            "upgradeStrategy": upgrade_strategy,
            "useQuota": use_quota,
            "timeout": timeout,
            "transferProtocol": transfer_protocol,
            "transferMethod": transfer_method
        }
        device_id = sanitize_request_param(device_id)
        firmware_resource_id = sanitize_request_param(firmware_resource_id)
        return self.session.post(self.get_url(f"/upgrade/{device_id}/{firmware_resource_id}"), params=params)

    @api_call(Task)
    def get_one(self, task_id: str) -> Response:
        task_id = sanitize_request_param(task_id)
        return self.session.get(self.get_url(f"/{task_id}"))

    @api_call_raw
    def update_one(self, task_id: str, config: TaskConfig):
        task_id = sanitize_request_param(task_id)
        return self.session.put(self.get_url(f"/{task_id}"), json=asdict(config))

    @api_call_raw
    def delete_one(self, task_id: str):
        task_id = sanitize_request_param(task_id)
        return self.session.delete(self.get_url(f"/{task_id}"))



from typing import Any, Optional
import requests

from .auth import Authenticator
from .v3.assigned_device_properties import AssignedDeviceProperties
from .v3.aws_integration import AwsIntegration
from .v3.cert_auth import CertAuth
from .v3.device_resources import DeviceResources
from .v3.device_tests import DeviceTests
from .v3.devices import Devices
from .v3.dialects import Dialects
from .v3.domains import Domains
from .v3.data_model import DataModel
from .v3.device_monitoring import DeviceMonitoring
from .v3.extensions import Extensions
from .v3.groups import Groups
from .v3.lifecycle_management import LifecycleManagement
from .v3.observations import Observations
from .v3.resources import Resources
from .v3.sessions import Sessions
from .v3.setting_values import SettingValues
from .v3.task_templates import TaskTemplates
from .v3.tasks import Tasks
from .v3.task_reports import TaskReports
from .device_client import DeviceClient
from .v3.users import Users


class Coiote:
    """
    Represents a connection to Coiote server.
    """
    SUPPORTED_API_VERSIONS = {"v3"}

    def __init__(
            self,
            url: str,
            auth: Optional[Any] = None,
            api_version: str = "v3",
            verify_ssl: bool = True
    ):
        self.url = url
        self.api_version = api_version
        if self.api_version not in Coiote.SUPPORTED_API_VERSIONS:
            raise ValueError("Only v3 API is supported right now")
        self.api_url = f"{url}/api/coiotedm/{api_version}"
        self.session = requests.Session()
        self.session.verify = verify_ssl
        self.authenticator = Authenticator(
            self.url, self.session, auth=auth)

        self.aws_integration: AwsIntegration = self._make_module(AwsIntegration)
        self.assigned_device_properties: AssignedDeviceProperties = self._make_module(AssignedDeviceProperties)
        self.cert_auth: CertAuth = self._make_module(CertAuth)
        self.data_model: DataModel = self._make_module(DataModel)
        self.device_monitoring: DeviceMonitoring = self._make_module(DeviceMonitoring)
        self.device_resources: DeviceResources = self._make_module(DeviceResources)
        self.device_tests: DeviceTests = self._make_module(DeviceTests)
        self.devices: Devices = self._make_module(Devices)
        self.dialects: Dialects = self._make_module(Dialects)
        self.domains: Domains = self._make_module(Domains)
        self.extensions: Extensions = self._make_module(Extensions)
        self.groups: Groups = self._make_module(Groups)
        self.lifecycle_management: LifecycleManagement = self._make_module(LifecycleManagement)
        self.observations: Observations = self._make_module(Observations)
        self.resources: Resources = self._make_module(Resources)
        self.sessions: Sessions = self._make_module(Sessions)
        self.setting_values: SettingValues = self._make_module(SettingValues)
        self.task_reports: TaskReports = self._make_module(TaskReports)
        self.task_templates: TaskTemplates = self._make_module(TaskTemplates)
        self.tasks: Tasks = self._make_module(Tasks)
        self.users: Users = self._make_module(Users)

    def _make_module(self, module_def):
        return module_def(root_url=self.api_url, authenticator=self.authenticator, session=self.session)

    def create_device_client(self, device_id: str) -> DeviceClient:
        return DeviceClient(device_id, self)



from dataclasses import dataclass
from enum import auto
from typing import List, Union

from coiote.utils import AutoNamedEnum


class InstanceType(AutoNamedEnum):
    Single = auto()
    Multiple = auto()


class Lwm2mResourceType(AutoNamedEnum):
    String = auto()
    Integer = auto()
    UnsignedInteger = auto()
    Float = auto()
    Boolean = auto()
    Opaque = auto()
    Time = auto()
    Objlnk = auto()
    Corelnk = auto()
    NoneType = "None"


class Lwm2mResourceOperationType(AutoNamedEnum):
    Read = auto()
    Write = auto()
    Execute = auto()


@dataclass
class Lwm2mResourceDefinition:
    id: int
    name: str
    operations: List[Lwm2mResourceOperationType]
    instanceType: InstanceType
    mandatory: bool
    type: Lwm2mResourceType
    range: str
    units: str
    description: str


@dataclass
class SingleInstanceResourceData:
    value: str


@dataclass
class MultiInstanceResourceData:
    values: List[str]


@dataclass
class SingleInstanceResourceValue:
    Single: SingleInstanceResourceData

    def get_value(self) -> str:
        return self.Single.value


@dataclass
class MultipleInstanceResourceValue:
    Multiple: MultiInstanceResourceData

    def get_value(self) -> List[str]:
        return self.Multiple.values


@dataclass
class Lwm2mResourceData:
    path: str
    value: Union[SingleInstanceResourceValue, MultipleInstanceResourceValue]
    resourceId: int
    operations: List[Lwm2mResourceOperationType]
    mandatory: bool
    type: Lwm2mResourceType
    range: str
    units: str
    description: str


@dataclass
class Lwm2mObjectDefinition:
    id: int
    name: str
    objectVersion: str
    instanceType: InstanceType
    mandatory: bool
    resourceDefinitions: List[Lwm2mResourceDefinition]



from typing import List

from requests import Response

from coiote.utils import ApiEndpoint, api_call, sanitize_request_param
from coiote.v3.model.device_resources import Lwm2mObjectDefinition, Lwm2mResourceData


class DeviceResources(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="instantiatedResources")

    @api_call(List[Lwm2mObjectDefinition])
    def get_defined_resources(self, device_id: str, objects: List[str]) -> Response:
        device_id = sanitize_request_param(device_id)
        parameters = {"parameters": objects}
        return self.session.get(self.get_url(f"/dataModelDefinition/{device_id}"), params=parameters)

    @api_call(List[Lwm2mResourceData])
    def get_resources_data(self, device_id: str, objects: List[str]) -> Response:
        device_id = sanitize_request_param(device_id)
        parameters = {"parameters": objects}
        return self.session.get(self.get_url(f"/resourcesData/{device_id}"), params=parameters)



from dataclasses import asdict

from requests import Response

from coiote.utils import ApiEndpoint, api_call, api_call_raw
from coiote.v3.model.task_templates import TaskTemplateInvocation
from coiote.v3.task_reports import TaskReport


class TaskTemplates(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="tasksFromTemplates")

    @api_call_raw
    def run_on_device(self, device_id: str, config: TaskTemplateInvocation, execute_immediately: bool) -> Response:
        return self.session.post(self.get_url(f"/device/{device_id}"),
                                 params={"executeImmediately": execute_immediately}, json=asdict(config))

    @api_call(TaskReport)
    def run_on_device_and_await(self, device_id: str, config: TaskTemplateInvocation,
                                wait_time_seconds: int) -> Response:
        return self.session.post(self.get_url(f"/device/{device_id}"),
                                 params={"waitTimeSeconds": f"{wait_time_seconds}s"}, json=asdict(config))

    @api_call_raw
    def run_on_group(self, group_id: str, config: TaskTemplateInvocation):
        return self.session.post(self.get_url(f"/group/{group_id}"), data=asdict(config))



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Dict, List


@dataclass
class AssignedDevicePropertyTarget:
    key: str
    value: str


@dataclass
class AssignedDeviceProperty:
    externalId: str
    creationTime: datetime
    isActive: bool
    wasApplied: bool
    lastApplication: Optional[datetime]
    properties: Dict[str, str]
    multiShot: bool
    lastMatchedDevice: Optional[str]
    target: AssignedDevicePropertyTarget


@dataclass
class AssignedDevicePropertyUpsertRequest:
    target: AssignedDevicePropertyTarget
    isActive: bool
    properties: Dict[str, str]
    multiShot: bool


@dataclass
class AssignedDevicePropertyUpdateRequest:
    isActive: bool
    multiShot: bool
    reapply: bool = False
    putProperties: Dict[str, str] = field(default_factory=dict)
    removeProperties: List[str] = field(default_factory=list)



from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Dict, Optional


@dataclass
class UserCreateRequest:
    login: str
    email: str
    password: str
    domain: str
    emailVerified: bool
    tosAccepted: bool
    roles: List[str] = field(default_factory=list)
    permissions: List[str] = field(default_factory=list)


@dataclass
class UserPatchRequest:
    emailVerified: Optional[bool] = None
    userEnabled: Optional[bool] = None
    domain: Optional[str] = None
    password: Optional[str] = None
    tosAccepted: Optional[bool] = None
    roles: List[str] = field(default_factory=list)
    permissions: List[str] = field(default_factory=list)


@dataclass
class User:
    userId: str
    login: str
    active: bool
    superUser: bool
    staff: bool
    domain: str
    password: str
    expirationDate: Optional[datetime]
    creationDate: Optional[datetime]
    roles: List[str]
    permissions: List[str]
    fullName: str
    properties: Dict[str, str]
    accessSchedule: Optional[str]
    securityPolicyId: Optional[str]



from dataclasses import dataclass
from typing import Union


@dataclass
class AzureDpsConfig:
    scopeId: str
    enrollmentGroupSymmetricKey: str
    dpsEndpoint: str


@dataclass
class AzureIotCentralConfig:
    apiToken: str
    iotCentralHost: str
    dpsHost: str
    syncEnabled: bool


@dataclass
class AzureIotHubConfig:
    connectionString: str
    storageConnectionString: str
    syncEnabled: bool


@dataclass
class AzureExtensionConfig:
    config: Union[AzureDpsConfig, AzureIotCentralConfig, AzureIotHubConfig]


@dataclass
class ExportConfig:
    groupId: str
    excludeIntegrated: bool = False



from dataclasses import dataclass, field
from enum import auto
from typing import Optional, Dict, List, Union

from coiote.utils import AutoNamedEnum


@dataclass
class InternalLocationData:
    fileName: str
    staticContent: bool = False


@dataclass
class ExternalLocationData:
    fileUrl: str
    username: Optional[str] = None
    password: Optional[str] = None


@dataclass
class InternalLocation:
    InternalLocation: InternalLocationData


@dataclass
class ExternalLocation:
    ExternalLocation: ExternalLocationData


class ResourceCategory(AutoNamedEnum):
    FIRMWARE = auto()
    SOFTWARE = auto()
    UNKNOWN = auto()
    IMAGE = auto()


class ResourceExpirationTime(AutoNamedEnum):
    ONE_DAY = auto()
    ONE_WEEK = auto()
    ONE_MONTH = auto()
    FOREVER = auto()


class DownloadProtocol(AutoNamedEnum):
    HTTP = auto()
    HTTPS = auto()
    COAP = auto()
    COAPS = auto()
    COAP_TCP = auto()
    COAP_TLS = auto()


@dataclass
class DownloadOptions:
    protocol: DownloadProtocol


@dataclass
class ResourceDownloadData:
    address: str


@dataclass
class Resource:
    name: str
    domain: str
    location: Union[InternalLocation, ExternalLocation]
    category: ResourceCategory
    expirationTime: ResourceExpirationTime
    visibleForSubtenants: bool = False
    description: Optional[str] = None
    id: Optional[str] = None
    device: Optional[str] = None
    properties: Dict[str, str] = field(default_factory=dict)
    directGroups: List[str] = field(default_factory=list)


@dataclass
class FileData:
    pass


@dataclass
class Base64FileData(FileData):
    data: str



from dataclasses import asdict

from coiote.utils import ApiEndpoint, api_call_raw
from coiote.v3.model.lifecycle_management import FactoryTestingRequest


class LifecycleManagement(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="deviceLifecycleManagement")

    @api_call_raw
    def activate_factory_testing(self, data: FactoryTestingRequest):
        return self.session.post(self.get_url("/factoryTesting/activate"), json=asdict(data))

    @api_call_raw
    def deactivate_factory_testing(self, data: FactoryTestingRequest):
        return self.session.post(self.get_url("/factoryTesting/deactivate"), json=asdict(data))



from dataclasses import asdict
from typing import List

from requests import Response

from coiote.utils import ApiEndpoint, api_call, sanitize_request_param, api_call_raw
from coiote.v3.model.extensions import AzureDpsConfig, AzureExtensionConfig, AzureIotHubConfig, AzureIotCentralConfig, \
    ExportConfig


class Extensions(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="extensions")

    @api_call(List[str])
    def get_available_extensions(self) -> Response:
        return self.session.get(self.get_url())

    @api_call(AzureDpsConfig)
    def get_azure_dps_config(self) -> Response:
        return self.session.get(self.get_url("/azureDps"))

    @api_call_raw
    def test_azure_dps_config(self, dps_config: AzureDpsConfig):
        return self.session.post(self.get_url("/azureDps/testConfig"), json=asdict(AzureExtensionConfig(dps_config)))

    @api_call_raw
    def setup_azure_dps_extension(self, dps_config: AzureDpsConfig):
        return self.session.post(self.get_url("/azureDps"), json=asdict(AzureExtensionConfig(dps_config)))

    @api_call_raw
    def delete_azure_dps_extension(self):
        return self.session.delete(self.get_url("/azureDps"))

    @api_call(AzureIotHubConfig)
    def get_azure_iot_hub_config(self) -> Response:
        return self.session.get(self.get_url("/azureIotHub"))

    @api_call_raw
    def test_azure_iot_hub_config(self, hub_config: AzureIotHubConfig):
        return self.session.post(self.get_url("/azureIotHub/testConfig"), json=asdict(AzureExtensionConfig(hub_config)))

    @api_call_raw
    def setup_azure_iot_hub_extension(self, hub_config: AzureIotHubConfig):
        return self.session.post(self.get_url("/azureIotHub"), json=asdict(AzureExtensionConfig(hub_config)))

    @api_call_raw
    def delete_azure_iot_hub_extension(self):
        return self.session.delete(self.get_url("/azureIotHub"))

    @api_call(AzureIotCentralConfig)
    def get_azure_iot_central_config(self) -> Response:
        return self.session.get(self.get_url("/azureIotCentral"))

    @api_call_raw
    def test_azure_iot_central_config(self, central_config: AzureIotCentralConfig):
        return self.session.post(self.get_url("/azureIotCentral/testConfig"),
                                 json=asdict(AzureExtensionConfig(central_config)))

    @api_call_raw
    def setup_azure_iot_central_extension(self, central_config: AzureIotCentralConfig):
        return self.session.post(self.get_url("/azureIotCentral"), json=asdict(AzureExtensionConfig(central_config)))

    @api_call_raw
    def delete_azure_iot_central_extension(self):
        return self.session.delete(self.get_url("/azureIotCentral"))

    @api_call(str)
    def export_group_to_hyperscaler(self, config: ExportConfig) -> Response:
        return self.session.post(self.get_url("/hyperscaler/export"), json=asdict(config))

    @api_call(str)
    def get_export_status(self, export_job_id: str) -> Response:
        export_job_id = sanitize_request_param(export_job_id)
        return self.session.get(self.get_url(f"/hyperscaler/export/{export_job_id}"))

    @api_call_raw
    def restart_device_client(self, device_id: str):
        device_id = sanitize_request_param(device_id)
        return self.session.get(self.get_url(f"/hyperscaler/restartClient/{device_id}"))



from dataclasses import asdict

from coiote.utils import ApiEndpoint, api_call_raw
from coiote.v3.model.cert_auth import CertificateData


class CertAuth(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="auth/certificates")

    @api_call_raw
    def add_certificate(self, certificate: CertificateData):
        return self.session.post(self.get_url(), json=asdict(certificate))

    @api_call_raw
    def delete_certificate(self):
        return self.session.delete(self.get_url())



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, Dict, List


@dataclass
class MonitoringStatus:
    enabled: bool


@dataclass
class Datapoint:
    date: datetime
    value: str


@dataclass
class MonitoringDataResponse:
    hasNext: bool
    lastPointTimestamp: Optional[datetime] = field(default=None)
    metadata: Dict[str, str] = field(default_factory=dict)
    data: List[Datapoint] = field(default_factory=list)

@dataclass
class SetResourceAliasRequest:
    lwm2mUrl: str
    alias: str


import base64
from dataclasses import asdict

from requests import Response

from coiote.utils import ApiEndpoint, api_call, sanitize_request_param, api_call_raw
from coiote.v3.model.resources import Resource, FileData, Base64FileData, DownloadOptions, \
    ResourceDownloadData


class Resources(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="resources")

    @api_call(str)
    def create_resource(self, resource: Resource) -> Response:
        print(asdict(resource))
        return self.session.post(self.get_url(), json=asdict(resource))

    @api_call_raw
    def upload_resource_data(self, resource_id: str, file: FileData):
        headers = {'Content-Type': 'application/octet-stream'}
        resource_id = sanitize_request_param(resource_id)
        if isinstance(file, Base64FileData):
            file_bytes = base64.b64decode(file.data)
            return self.session.put(self.get_url(f"/{resource_id}/data"), data=file_bytes, headers=headers)
        else:
            raise ValueError("Invalid file type")

    @api_call(ResourceDownloadData)
    def get_resource_download_url(self, resource_id: str, options: DownloadOptions) -> Response:
        resource_id = sanitize_request_param(resource_id)
        return self.session.post(self.get_url(f"/{resource_id}/downloadUrl"), json=asdict(options))



from dataclasses import dataclass, field
from datetime import datetime
from enum import auto
from typing import Optional, Dict, List, Union

from coiote.utils import AutoNamedEnum


@dataclass
class PlaintextPsk:
    PlainTextPsk: str


@dataclass
class BinaryPsk:
    BinaryPsk: List[int]


@dataclass
class HexStringPsk:
    HexadecimalPsk: str


DtlsPreSharedKey = Union[PlaintextPsk, BinaryPsk, HexStringPsk]


class ConnectorType(AutoNamedEnum):
    management = auto()
    bootstrap = auto()


@dataclass
class Device:
    domain: Optional[str] = None
    id: Optional[str] = None
    lastSessionTime: Optional[datetime] = None
    lastContactTime: Optional[datetime] = None
    lastRegisterTime: Optional[datetime] = None
    firstRegisterTime: Optional[datetime] = None
    creationTime: Optional[datetime] = None
    ipAddress: Optional[str] = None
    serialNumber: Optional[str] = None
    oui: Optional[str] = None
    modelName: Optional[str] = None
    hardwareVersion: Optional[str] = None
    softwareVersion: Optional[str] = None
    productClass: Optional[str] = None
    manufacturer: Optional[str] = None
    description: Optional[str] = None
    friendlyName: Optional[str] = None
    securityMode: Optional[str] = None
    dtlsIdentity: Optional[str] = None
    dtlsPsk: Optional[DtlsPreSharedKey] = None
    connectorType: ConnectorType = ConnectorType.management
    bootstrap: bool = False
    directGroups: List[str] = field(default_factory=list)
    properties: Dict[str, str] = field(default_factory=dict)
    blacklisted: bool = False
    managementEnabled: bool = False


@dataclass
class DevicesFindResponse:
    devices: List[Device] = field(default_factory=list)
    nextPageBookmark: Optional[str] = None


@dataclass
class DeviceUpdateRequest:
    addProperties: Dict[str, str] = field(default_factory=dict)
    removeProperties: List[str] = field(default_factory=list)
    addGroups: List[str] = field(default_factory=list)
    removeGroups: List[str] = field(default_factory=list)
    blacklisted: bool = False
    managementEnabled: bool = True
    description: Optional[str] = None
    friendlyName: Optional[str] = None
    securityMode: Optional[str] = None
    dtlsIdentity: Optional[str] = None
    dtlsPsk: Optional[DtlsPreSharedKey] = None


@dataclass
class SingleDeviceAddResponse:
    id: str
    result: str


@dataclass
class DevicesAddResponse:
    id: str
    success: bool
    reason: Optional[str] = None


@dataclass
class DevicesBatchAddResponse:
    total: int
    succeeded: int
    failed: int
    results: List[DevicesAddResponse]



from dataclasses import asdict
from typing import Optional, List

from requests import Response

from coiote.utils import ApiEndpoint, api_call, sanitize_request_param, api_call_raw
from coiote.v3.model.users import UserCreateRequest, User, UserPatchRequest


class Users(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="users")

    @api_call(List[str])
    def get_all(self, search_criteria: Optional[str] = None) -> Response:
        return self.session.get(self.get_url(), params={"searchCriteria": search_criteria})

    @api_call_raw
    def create_one(self, user: UserCreateRequest):
        return self.session.post(self.get_url(), json=asdict(user))

    @api_call(List[str])
    def find_many(self, username: Optional[str] = None, email: Optional[str] = None, domain: Optional[str] = None,
                  active_only: bool = False):
        params = {
            "username": username,
            "email": email,
            "domain": domain,
            "isActiveOnly": active_only
        }
        return self.session.get(self.get_url("/search/usernames"), params=params)

    @api_call(User)
    def get_one(self, username: str):
        username = sanitize_request_param(username)
        return self.session.get(self.get_url(f"/{username}"))

    @api_call_raw
    def delete_one(self, username: str):
        username = sanitize_request_param(username)
        return self.session.delete(self.get_url(f"/{username}"))

    @api_call_raw
    def update_one(self, username: str, user: UserPatchRequest):
        username = sanitize_request_param(username)
        return self.session.patch(self.get_url(f"/{username}"), json=asdict(user))



from dataclasses import dataclass, field
from enum import auto
from typing import Optional, List, Dict, Union

from coiote.utils import AutoNamedEnum


@dataclass
class ReadDefinition:
    key: str


@dataclass
class ReadOperation:
    read: ReadDefinition


@dataclass
class WriteDefinition:
    key: str
    value: str


@dataclass
class WriteOperation:
    write: WriteDefinition


@dataclass
class ExecuteArg:
    digit: int
    argument: Optional[str] = None


@dataclass
class ExecuteDefinition:
    key: str
    argumentList: List[ExecuteArg] = field(default_factory=list)


@dataclass
class ExecuteOperation:
    execute: ExecuteDefinition


DeviceOperation = Union[ReadOperation, WriteOperation, ExecuteOperation]


@dataclass
class ConfigurationTaskDefinition:
    name: str
    batchRequests: bool = True
    executeImmediately: bool = True
    operations: List[DeviceOperation] = field(default_factory=list)


class TransferMethod(AutoNamedEnum):
    Pull = auto()
    Push = auto()


class TransferProtocol(AutoNamedEnum):
    HTTP = auto()
    HTTPS = auto()
    COAP = auto()
    COAPS = auto()
    COAP_TCP = auto()
    COAP_TLS = auto()


class UpgradeStrategy(AutoNamedEnum):
    ObservationTrigger = auto()
    WithoutObservations = auto()
    ObservationBased = auto()
    SendBased = auto()


@dataclass
class TaskParameter:
    name: str
    value: str


@dataclass
class TaskConfig:
    taskName: str
    parameters: List[TaskParameter]
    properties: Dict[str, str]
    isActive: Optional[bool] = False


@dataclass
class Task:
    id: str
    config: TaskConfig



from dataclasses import asdict
from typing import List, Optional

from requests import Response

from coiote.utils import ApiEndpoint, api_call, sanitize_request_param, api_call_raw
from coiote.v3.model.observations import ObservationData, EntityObservationData, SetObservationRequest, \
    BasicSetObservationRequest


class Observations(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="observations")

    @api_call(List[EntityObservationData])
    def get_all(self, search_criteria: Optional[str] = None) -> Response:
        return self.session.get(self.get_url(), params={"searchCriteria": search_criteria})

    @api_call(ObservationData)
    def get_device_observation_data(self, device_id: str, path: str) -> Response:
        device_id = sanitize_request_param(device_id)
        path = sanitize_request_param(path)
        return self.session.get(self.get_url(f"/device/{device_id}/{path}"))

    @api_call_raw
    def set_observation_on_device(self, device_id: str, path: str, data: SetObservationRequest):
        device_id = sanitize_request_param(device_id)
        path = sanitize_request_param(path)
        return self.session.post(self.get_url(f"/device/{device_id}/{path}"), json=asdict(data))

    @api_call_raw
    def delete_observation_for_device(self, device_id: str, path: str):
        device_id = sanitize_request_param(device_id)
        path = sanitize_request_param(path)
        return self.session.delete(self.get_url(f"/device/{device_id}/{path}"))

    @api_call_raw
    def set_observation_on_group(self, group_id: str, path: Optional[str] = None, lwm2m_url: Optional[str] = None,
                                 data: BasicSetObservationRequest = BasicSetObservationRequest()):
        group_id = sanitize_request_param(group_id)
        if path is not None:
            path = sanitize_request_param(path)
            api_path = f"/group/resourcePath/{group_id}/{path}"
        elif lwm2m_url is not None:
            lwm2m_url = sanitize_request_param(lwm2m_url)
            api_path = f"/group/resourceUrl/{group_id}/{lwm2m_url}"
        else:
            raise ValueError("You must specify either datamodel path or LwM2M Url for the observation")
        return self.session.post(self.get_url(api_path), json=asdict(data))



from dataclasses import asdict
from typing import List, Optional

from requests import Response

from coiote.utils import ApiEndpoint, StringResult, api_call, sanitize_request_param, api_call_raw
from coiote.v3.model.assigned_device_properties import AssignedDeviceProperty, AssignedDevicePropertyUpsertRequest, \
    AssignedDevicePropertyUpdateRequest


class AssignedDeviceProperties(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="assignedDeviceProperties")

    @api_call(List[str])
    def get_all(self, search_criteria: Optional[str] = None) -> Response:
        return self.session.get(self.get_url(), params={"searchCriteria": search_criteria})

    @api_call(StringResult)
    def upsert_property(self, property_id: str, property: AssignedDevicePropertyUpsertRequest) -> Response:
        property_id = sanitize_request_param(property_id)
        return self.session.put(self.get_url(f"/{property_id}"), json=asdict(property))

    @api_call(AssignedDeviceProperty)
    def get_one(self, property_id: str) -> Response:
        property_id = sanitize_request_param(property_id)
        return self.session.get(self.get_url(f"/{property_id}"))

    @api_call_raw
    def delete_one(self, property_id: str) -> Response:
        property_id = sanitize_request_param(property_id)
        return self.session.delete(self.get_url(f"/{property_id}"))

    @api_call(AssignedDeviceProperty)
    def update_one(self, property_id: str, update: AssignedDevicePropertyUpdateRequest) -> Response:
        property_id = sanitize_request_param(property_id)
        return self.session.patch(self.get_url(f"/{property_id}"), json=asdict(update))



from datetime import datetime
from typing import List, Optional

from requests import Response

from coiote.utils import ApiEndpoint, api_call, ISO_INSTANT_FORMAT, sanitize_request_param
from coiote.v3.model.task_reports import ExecutedTaskInfo, TaskReportBatch, TasksSummary, TaskReport


class TaskReports(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="taskReports")

    @api_call(List[ExecutedTaskInfo])
    def get_reports(self, query: str) -> Response:
        return self.session.get(self.get_url(""), params={"searchCriteria": query})

    @api_call(int)
    def get_reports_cursor(self, query: str) -> Response:
        return self.session.get(self.get_url("/findReports"), params={"searchCriteria": query})

    @api_call(TaskReportBatch)
    def get_next_reports_batch(self, cursor_id: int, count: int) -> Response:
        return self.session.get(self.get_url("/moreReports"), params={"cursor": cursor_id, "count": count})

    @api_call(List[TaskReport])
    def find_reports_since(self,
                           since: datetime,
                           taskId: Optional[str] = None,
                           device_id: Optional[str] = None,
                           limit: int = 512,
                           polling_interval_seconds: Optional[int] = None,
                           only_direct_tasks: bool = False) -> Response:
        params = {
            "lastUpdateTime": since.strftime(ISO_INSTANT_FORMAT),
            "taskId": taskId,
            "deviceId": device_id,
            "limit": limit,
            "pollingInterval": polling_interval_seconds,
            "directTasksOnly": only_direct_tasks
        }
        return self.session.get(self.get_url("/findSince"), params=params)

    @api_call(TasksSummary)
    def get_task_summary(self, task_id: str) -> Response:
        return self.session.get(self.get_url("/summary"), params={"taskId": task_id})

    @api_call(TaskReport)
    def get_report_for_device_task(self, task_id: str, device_id: str) -> Response:
        task_id = sanitize_request_param(task_id)
        device_id = sanitize_request_param(device_id)
        return self.session.get(self.get_url(f"/{task_id}/{device_id}"))



from dataclasses import dataclass
from typing import Optional


@dataclass
class Group:
    id: str
    description: Optional[str] = None


@dataclass
class GroupUpdateRequest:
    description: Optional[str]



from dataclasses import asdict
from typing import List, Dict

from requests import Response

from coiote.utils import ApiEndpoint, api_call, sanitize_request_param, api_call_raw
from coiote.v3.model.device_tests import DeviceTestSchedule, TestExecutionReport


class DeviceTests(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="protocolTests")

    @api_call_raw
    def run_for_device(self, device_id: str, tests: List[str] = None):
        if tests is None:
            tests = []
        device_id = sanitize_request_param(device_id)
        tests = DeviceTestSchedule(tests)
        return self.session.post(self.get_url(f"/schedule/device/{device_id}"), json=asdict(tests))

    @api_call_raw
    def run_for_group(self, group_id: str, tests: List[str] = None):
        if tests is None:
            tests = []
        group_id = sanitize_request_param(group_id)
        tests = DeviceTestSchedule(tests)
        return self.session.post(self.get_url(f"/schedule/group/{group_id}"), json=asdict(tests))

    @api_call(TestExecutionReport)
    def get_report_for_device(self, device_id: str, tests: List[str] = None) -> Response:
        if tests is None:
            tests = []
        device_id = sanitize_request_param(device_id)
        tests = DeviceTestSchedule(tests)
        return self.session.post(self.get_url(f"/report/device/{device_id}"), json=asdict(tests))

    @api_call(Dict[str, TestExecutionReport])
    def get_report_for_group(self, group_id: str, tests: List[str] = None) -> Response:
        if tests is None:
            tests = []
        group_id = sanitize_request_param(group_id)
        return self.session.post(self.get_url(f"/report/group/{group_id}"), json=asdict(DeviceTestSchedule(tests)))



from dataclasses import dataclass, field
from typing import Optional, Union, List


@dataclass
class ObservedDevice:
    DeviceId: str


@dataclass
class ObservedGroup:
    GroupId: str


@dataclass
class EntityObservationData:
    targetEntity: Union[ObservedDevice, ObservedGroup]
    path: str


@dataclass
class ObservationAttribute:
    name: str
    value: str


@dataclass
class ObservationData:
    deviceId: str
    path: str
    attributes: List[ObservationAttribute] = field(default_factory=list)


@dataclass
class SetObservationRequest:
    attributes: List[ObservationAttribute] = field(default_factory=list)
    createEnsureObserveIfNotExists: bool = False


@dataclass
class BasicSetObservationRequest:
    minPeriod: Optional[int] = None
    maxPeriod: Optional[int] = None
    createEnsureObserveIfNotExists: bool = False



from dataclasses import asdict

from coiote.utils import ApiEndpoint, api_call, api_call_raw
from coiote.v3.model.aws_integration import AwsCertificateData


class AwsIntegration(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="awsIntegration")

    @api_call_raw
    def add_certificate(self, certificate: AwsCertificateData):
        return self.session.post(self.get_url("/auth/externalCertificate"), json=asdict(certificate))

    @api_call_raw
    def delete_certificate(self):
        return self.session.delete(self.get_url("/auth/externalCertificate"))



from dataclasses import asdict
from typing import List, Optional

from requests import Response

from coiote.utils import ApiEndpoint, StringResult, api_call, sanitize_request_param, api_call_raw
from coiote.v3.model.groups import GroupUpdateRequest, Group


class Groups(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="groups")

    @api_call(List[str])
    def get_all(self, search_criteria: Optional[str] = None) -> Response:
        return self.session.get(self.get_url(), params={"searchCriteria": search_criteria})

    @api_call(StringResult)
    def add_group(self, group: Group) -> Response:
        return self.session.post(self.get_url(), json=asdict(group))

    @api_call(Group)
    def get_one(self, group_id: str) -> Response:
        group_id = sanitize_request_param(group_id)
        return self.session.get(self.get_url(f"/{group_id}"))

    @api_call_raw
    def update_one(self, group_id: str, update: GroupUpdateRequest):
        group_id = sanitize_request_param(group_id)
        return self.session.put(self.get_url(f"/{group_id}"), json=asdict(update))

    @api_call_raw
    def delete_one(self, group_id: str):
        group_id = sanitize_request_param(group_id)
        return self.session.delete(self.get_url(f"/{group_id}"))



from typing import Optional

from coiote.utils import ApiEndpoint, sanitize_request_param, api_call_raw


class Sessions(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="sessions")

    @api_call_raw
    def start_for_device(self, device_id: str, task_id: Optional[str] = None, allow_deregistered=False):
        device_id = sanitize_request_param(device_id)
        params = {"taskId": task_id}
        if allow_deregistered:
            path = f"/{device_id}/allow-deregistered"
        else:
            path = f"/{device_id}"
        return self.session.post(self.get_url(path), params=params)



import functools
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, List, Optional, Dict, Type, Callable, TypeVar, get_args, Union
from urllib.parse import quote as sanitize_param

from dacite import from_dict, config
from dateutil import parser
from requests import Response, Session
from typing_extensions import ParamSpec, get_origin

from coiote.auth import Authenticator

TYPE_HOOKS: Dict[Type[Any], Callable[[Any], Any]] = {
    datetime: parser.isoparse
}

ISO_INSTANT_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

T = TypeVar('T')
P = ParamSpec('P')


@dataclass
class StringResult:
    result: str


class ApiEndpoint:
    def __init__(self,
                 root_url: str,
                 api_url: str,
                 authenticator: Authenticator,
                 session: Session):
        self.root_url = root_url
        self.authenticator = authenticator
        self.session = session
        self.api_url = api_url

    def get_url(self, endpoint: str = "") -> str:
        return f"{self.root_url}/{self.api_url}{endpoint}"


def deserialize_one(json, response_type: Type[T]) -> T:
    if response_type is int:
        return int(json)
    elif response_type is str:
        return str(json)
    elif response_type is float:
        return float(json)
    else:
        return from_dict(data_class=response_type, data=json, config=config.Config(type_hooks=TYPE_HOOKS, cast=[Enum]))


def _safe_get_json(response: Optional[Response]) -> Any:
    try:
        return response.json() if response is not None else None
    except:
        return None


def _get_response_summary(response: Response):
    json = _safe_get_json(response)
    headers = response.headers
    code = response.status_code
    return f"Code:\n{code}\nHeaders:\n{headers}\nBody:\n{json}"


def _make_error(msg: str, response: Optional[Response]):
    if response:
        summary = _get_response_summary(response)
    else:
        summary = "No response data."
    return ValueError(f"{msg}\n{summary}")


def is_optional(field: Type[T]) -> bool:
    return get_origin(field) is Union and type(None) in get_args(field)


def deserialize_response(response_body: Optional[Any], expected_type: Type[T]) -> T:
    if is_optional(expected_type):
        class_tag = get_args(expected_type)[0]
        return deserialize_one(response_body, class_tag) if response_body is not None else None
    elif hasattr(expected_type, '__origin__') and expected_type.__origin__ in {list, List}:
        class_tag = get_args(expected_type)[0]
        return [deserialize_one(element, response_type=class_tag) for element in response_body]
    elif hasattr(expected_type, '__origin__') and expected_type.__origin__ in {dict, Dict}:
        key_tag, value_tag = get_args(expected_type)
        return {deserialize_one(key, response_type=key_tag): deserialize_one(value, response_type=value_tag) for
                key, value in
                response_body.items()}
    else:
        return deserialize_one(response_body, response_type=expected_type)


def handle_response(response: Response, expected_type: Type[T]) -> T:
    if response is not None:
        if response.ok:
            return deserialize_response(response.json(), expected_type)
        else:
            raise _make_error("Received error response from the server", response)
    else:
        raise ValueError("Failed to get the server response, most likely due to connection errors.")


def api_call_raw(func: Callable[P, Response]) -> Callable[P, Response]:
    @functools.wraps(func)
    def wrap_with_response_handle(*args: P.args, **kwargs: P.kwargs) -> Response:
        api_endpoint, *rest = args
        api_endpoint.authenticator.authenticate()  # type: ignore
        response = func(*args, **kwargs)
        return response

    return wrap_with_response_handle


def api_call(response_format: Type[T]) -> Callable[[Callable[P, Response]], Callable[P, T]]:
    def decorator_deserialize(func: Callable[P, Response]) -> Callable[P, T]:
        @functools.wraps(func)
        def wrap_with_auth_and_deserialization(*args: P.args, **kwargs: P.kwargs) -> T:
            api_endpoint, *rest = args
            api_endpoint.authenticator.authenticate()  # type: ignore
            response = func(*args, **kwargs)
            return handle_response(response, response_format)

        return wrap_with_auth_and_deserialization

    return decorator_deserialize


def sanitize_request_param(param: Optional[str]) -> Optional[str]:
    if param is None:
        return None
    else:
        return sanitize_param(param, safe='')


class AutoNamedEnum(Enum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name



from dataclasses import dataclass
from typing import Optional, List, Dict


@dataclass
class TaskTemplateConfig:
    taskName: str
    parameters: List[Dict[str, str]]
    isActive: bool
    properties: Dict[str, str]


@dataclass
class TaskTemplateInvocation:
    templateName: str
    config: Optional[TaskTemplateConfig] = None
    callback: Optional[dict] = None
